#!/usr/bin/env tclsh

#******************************
# 
# osDubletten
# Version 2016-08-12
# Copyright (C) 2016 Oliver Scholl < tcltk@posteo.de > < www.tcltk.info >
#
# Operating System:
# Linux
#
# Requirement:
# tcl 8.5 or higher
# tk 8.5 or higher
# md5sum
#
# Description of the program:
# This program searches for duplicate files and writes them into a text file.
# It works hierarchical:
# First it checks the file size and keeps the files with the same file size,
# then it checks severals bytes of the files and keeps the files with the same bytes 
# and at last it calculates the md5sum of the remaining files and keeps the files with the same md5sum.
# This process saves much time compared with the time to calculate md5sum of all files.
# 
# How to start the program from the shell:
# tclsh /Path/To/osDubletten.tcl /Path/To/The/Directory/Where/To/Start
# 
# License:
# This program is free software; you can redistribute it and/or 
# modify it under the terms of the GNU General Public License as 
# published by the Free Software Foundation; either version 2 of 
# the license, or any later version.
#
# This program is distributed in the hope that it will be useful, 
# but WITHOUT ANY WARRANTY; without even the implied warranty of 
# merchantability or fitness for a particular purpose. 
# See the GNU General Public License for mor details.
#
#*************************************

package require fileutil
#package require md5

if {$argv eq ""} {
	puts "Es wurde kein Startordner angegeben."
	exit
}

set Ordner [lindex $argv 0]
set Dateien {} ; # Liste mit allen Dateien
set Ausgabedatei [file join $env(HOME) Dubletten.txt]

proc Vergleich {Dateiname} {
	if {[file type $Dateiname] == "file"} {
		return 1
	} else {
		return 0
	}
}

proc DublettenFinden {Dateien ArtDerRueckgabe} {
	# Selektiert die Dubletten aus der Liste der Dateien
	# Der Fingerabdruck ist durch ein Leerzeichen vom Dateinamen abgetrennt
	# Dateien: Liste der Dateien
	# ArtDerRueckgabe: 1=nur Dateiname, 2=Fingerabdruck und Dateiname
	# Rückgabe: Liste der Dubletten

	set Dateien [lsort -unique $Dateien]

	set Dubletten {}
	for {set Index 0} {$Index < [expr [llength $Dateien]-1]} {incr Index} {
		set Element1 [lindex $Dateien $Index]
		set tmpIndex [string first " " $Element1]
		set Fingerabdruck1 [string range $Element1 0 $tmpIndex]
		incr tmpIndex
		set Datei1 [string range $Element1 $tmpIndex end]

		set Element2 [lindex $Dateien [expr $Index+1]]
		set tmpIndex [string first " " $Element2]
		set Fingerabdruck2 [string range $Element2 0 $tmpIndex]
		incr tmpIndex
		set Datei2 [string range $Element2 $tmpIndex end]

		if {$Fingerabdruck1 == $Fingerabdruck2} {
			if {$ArtDerRueckgabe == 1} {
				lappend Dubletten $Datei1
				lappend Dubletten $Datei2
			} else {
				lappend Dubletten $Element1
				lappend Dubletten $Element2
			}
		}
	}
	set Dubletten [lsort -unique $Dubletten]
	return $Dubletten
}

proc BinaerPruefen {Dateien Anzahl} {
	# Extrahiert einige Bytes aus den Dateien als Fingerabdruck
	# Dateien: Liste der Dateien
	# Anzahl: Anzahl der Stellen, an denen geprüft werden soll
	incr Anzahl ; # sonst wird eine Stelle zu wenig genommen
	for {set Index 0} {$Index < [llength $Dateien]} {incr Index} {
		set Datei [lindex $Dateien $Index]
		set Groesse [file size $Datei]
		set fDatei [open $Datei r]
		fconfigure $fDatei -translation binary
		
		set tmpDatei ""
		set Schrittweite [expr $Groesse/$Anzahl]
		for {set Index2 1} {$Index2 < $Anzahl} {incr Index2} {
			set Position [expr $Schrittweite*$Index2]
			seek $fDatei $Position start
			set Zeichen [read $fDatei 1]
			set Dezimal [scan $Zeichen %c]
			append tmpDatei $Dezimal
			append tmpDatei "-"
		}
		close $fDatei
		set Element "$tmpDatei $Datei"
		set Dateien [lreplace $Dateien $Index $Index $Element]
	}
	return $Dateien
}

puts "Dateien suchen ..."
for {set Index 0} {$Index < [llength $Ordner]} {incr Index} {
	set tmpDateien [fileutil::find [lindex $Ordner $Index] Vergleich]
	set Dateien [concat $Dateien $tmpDateien]
}
puts "fertig. [llength $Dateien] Dateien."

# Größe bestimmen
puts "Dateien mit gleicher Groesse suchen ..."
for {set Index 0} {$Index < [llength $Dateien]} {incr Index} {
	set Datei [lindex $Dateien $Index]
	set Groesse [file size $Datei]
	set Element "$Groesse $Datei"
	set Dateien [lreplace $Dateien $Index $Index $Element]
}
set Dateien [DublettenFinden $Dateien 1]
puts "fertig. Noch [llength $Dateien] Dateien."

# Binär einlesen und einige Zeichen extrahieren
puts "Dateien mit gleichem Bytecode an 4 Stellen suchen ..."
set Dateien [BinaerPruefen $Dateien 4]
set Dateien [DublettenFinden $Dateien 1]
puts "fertig. Noch [llength $Dateien] Dateien."

# Binär einlesen und einige Zeichen extrahieren
puts "Dateien mit gleichem Bytecode an 10 Stellen suchen ..."
set Dateien [BinaerPruefen $Dateien 10]
set Dateien [DublettenFinden $Dateien 1]
puts "fertig. Noch [llength $Dateien] Dateien."

# md5sum berechnen
puts "Dateien mit gleicher md5sum suchen ..."
for {set Index 0} {$Index < [llength $Dateien]} {incr Index} {
	set Datei [lindex $Dateien $Index]
	set Md5sum ""
	catch {
		#set Md5sum [md5::md5 -hex -file $Datei]
		set Md5sum [exec md5sum $Datei]
	}
	#set Element "$Md5sum $Datei"
	set Element $Md5sum
	set Dateien [lreplace $Dateien $Index $Index $Element]
}
set Dateien [DublettenFinden $Dateien 2]
puts "fertig. Es gibt [llength $Dateien] Dubletten."

if {[llength $Dateien] > 0} {
	set f [open $Ausgabedatei w]
	foreach Element $Dateien {
		puts $f $Element
	}
	close $f
	puts "Die Dubletten wurden in folgender Datei gespeichert:"
	puts $Ausgabedatei
} else {
	puts "Es gibt keine Dubletten."
}
